from flask import Flask

from highfield import loading
from highfield.defaults import *
from config import *

class Application(Flask):

    def __init__(self):
        template_folder = loading.path('views')
        static_folder = loading.path('static')
        super(Application, self).__init__(__name__,
                                          template_folder=template_folder,
                                          static_folder=static_folder)
        self.configure()
        self.session_cookie_name = session_cookie_name
        self.session_secret_key = session_secret_key
        self.csrf_token_lifespan = csrf_token_lifespan
        pass

    def configure(self):
        self.controllers = loading.namespace('controllers')
        self.models = loading.namespace('models')

        for path, route in routes:
            c_name, a_name = route.get('to').split('.')
            methods = [m.upper() for m in route.get('via')]

            def closure(c_name, a_name):
                def dispatch(*args, **kwargs):
                    controller = getattr(self.controllers, c_name)(self, a_name)
                    return getattr(controller, a_name).__call__(*args, **kwargs)
                return dispatch
            self.add_url_rule(path, route.get('to'), closure(c_name, a_name), methods=methods)
            pass
        return self

    def run(self, network=False, **kwargs):
        if network:
            kwargs.update(host='0.0.0.0')
            pass
        super(Application, self).run(**kwargs)
        pass
    pass




from highfield import loading
import os

def mkpath(path, module=False):
    try:
        os.makedirs(path)
        if module:
            mkfile(os.path.join(path, '__init__.py'), '')
        pass
    except OSError as e:
        if not os.path.isdir(path):
            raise
        pass
    pass

def mkfile(filename, content=''):
    with open(filename, 'w') as _file:
        _file.write(content)

def init():
    mkpath(loading.path('controllers'), module=True)
    mkpath(loading.path('models'), module=True)
    mkpath(loading.path('helpers'), module=True)
    mkpath(loading.path('views'))
    mkpath(loading.path('static/css/sources'))
    mkfile(loading.path('static/css/application.css'))
    mkpath(loading.path('static/js/'))
    mkfile(loading.path('static/js/application.js'))


    config = ("mongodb_uri = None\n"
              "db_name = None\n\n"
              "routes = [('/static/<path:filepath>',\n"
              "           {'to': 'static.index',\n"
              "            'via': ['get']})]\n"
             )
    mkfile(loading.path('config.py'), config)

    static_controller = ("from highfield.base_controller import BaseController\n"
                         "\n"
                         "class StaticController(BaseController):\n"
                         "    def index(self, filepath):\n"
                         "        return self.static_file(filepath)\n"
                         "    pass\n"
                        )
    mkfile(loading.path('controllers/static_controller.py'), static_controller)

    server = ("from highfield.application import Application\n\n"
              "Application().run(network=True, debug=True)")
    mkfile(loading.path('server.py'), server)
    pass



from highfield import naming

import pkgutil
import os

def path(folder_name):
    return os.path.join(os.getcwd(), folder_name)

def namespace(module_name):
    module_path = path(module_name)
    temp = __import__(module_name, globals(), locals(), [], -1)
    ns = type('Namespace', (object,), {})
    for name, component in generate_component(module_name, module_path):
        setattr(component, 'namespace', ns)
        setattr(ns, name, component)
        pass
    return ns

def generate_component(module_name, module_path):
    component_type_name = naming.singular(module_name)
    for loader, submodule_name in generate_submodule_loader(module_path):
        if not submodule_name.endswith(component_type_name):
            continue
            pass

        sm = submodule(loader, module_name, submodule_name)
        expected_name = naming.file_to_class(submodule_name)
        if expected_name in sm.__dict__:
            yield (naming.file_to_canonical(submodule_name),
                   sm.__dict__[expected_name])
        pass
    pass

def submodule(loader, module_name, submodule_name):
    full_module_name = '%s.%s' % (module_name, submodule_name)
    return loader.find_module(submodule_name).load_module(full_module_name)

def generate_submodule_loader(module_path):
    for loader, submodule_name, tmp in pkgutil.iter_modules([module_path]):
        yield (loader, submodule_name)
        pass
    pass



from pymongo import MongoClient

from highfield import naming
from highfield.errors import *
from highfield.defaults import *
from config import *

mongo = MongoClient(mongodb_uri)

class MetaModel(type):
    def __init__(cls, name, bases, clsdict):
        cls.canonical_name = naming.class_to_canonical(name)
        cls.human_name = naming.canonical_to_words(cls.canonical_name)
        cls.collection = getattr(getattr(mongo, db_name), cls.canonical_name)
        cls.properties = type('ModelProperties', (object,), {})
        cls.set_properties()
        super(MetaModel, cls).__init__(name, bases, clsdict)
        pass
    pass

class BaseModel(object):
    __metaclass__ = MetaModel

    class Property(object):
        def __init__(self, canonical_name, validator=None, *args, **kwargs):
            self.canonical_name = canonical_name
            self.human_name = naming.canonical_to_words(self.canonical_name)
            self.validator = validator
            for purpose, required in kwargs.iteritems():
                setattr(self, purpose, True if required else False)
            pass

        def valid(self, value):
            if not self.validator:
                return True
            elif hasattr(self.validator, 'match'):
                if self.validator.match(value):
                    return True
                pass
            elif hasattr(self.validator, '__call__'):
                if self.validator(value):
                    return True
                pass
            return False
        pass

    @classmethod
    def new_property(cls, canonical_name, validator=None, *args, **kwargs):
        setattr(cls.properties, canonical_name, cls.Property(canonical_name, validator, *args, **kwargs))
        pass

    @classmethod
    def set_properties(cls):
        raise NotImplementedError()

    @classmethod
    def has_required_properties(cls, purpose, doc):
        if not doc:
            return False
        for k, v in cls.properties.__dict__.iteritems():
            if isinstance(v, cls.Property) and getattr(v, purpose, False):
                if not v in doc:
                    return False
                pass
            pass
        return True

    @classmethod
    def extract_properties(cls, doc):
        result = {}
        for k, v in cls.properties.__dict__.iteritems():
            if isinstance(v, cls.Property) and k in doc:
                result[k] = doc[k]
                pass
            pass
        return result

    @classmethod
    def such(cls, that, limit=25, page=1):
        skip = limit * (page - 1)
        return [item for item in cls.collection.find(that,
                                                     skip=skip,
                                                     limit=limit)]

    @classmethod
    def every(cls, max=25, page=1):
        return cls.such({}, limit=limit, page=page)

    @classmethod
    def one(cls, **kwargs):
        return cls.collection.find_one(kwargs)

    @classmethod
    def update(cls, query, update, upsert=True, new=False):
        return cls.collection.find_and_modify(query=query,
                                              update=update,
                                              upsert=upsert,
                                              new=new)

    @classmethod
    def new(cls, doc):
        return cls.collection.insert(doc)

    @classmethod
    def exists(cls, **kwargs):
        return True if cls.one(**kwargs) else False

    @classmethod
    def existing(cls, **kwargs):
        existing = cls.one(**kwargs)
        if not existing:
            properties = naming.enumeration([getattr(cls.properties, key).human_name for key in kwargs.keys()])
            string = 'Could not find a %s with the given %s' % (cls.human_name, properties)
            raise ValidationError({cls.canonical_name: string})
            pass
        return existing

    @classmethod
    def validates(cls, **kwargs):
        for (k, v) in kwargs.iteritems():
            if not getattr(cls.properties, k).valid(v):
                return False
            pass
        return True

    @classmethod
    def validate(cls, **kwargs):
        invalid = []
        for (k, v) in kwargs.iteritems():
            if not getattr(cls.properties, k).valid(v):
                invalid.append(cls.properties[k].human_name)
                pass
            pass
        if len(invalid):
            raise ValidationError({cls.canonical_name: 'Invalid %s' % naming.enumeration(invalid)})
            pass
        return True

    @classmethod
    def available(cls, **kwargs):
        return not cls.exists(**kwargs)
    pass



from highfield.helpers.random import unique_id
from highfield.errors import *

from datetime import datetime as dt

import time

def original(decorator, *args, **kwargs):
    def wrapper(*args, **kwargs):
        wrapped = decorator(*args, **kwargs)
        if hasattr(args[0], 'original'):
            wrapped.original = args[0].original
            pass
        else:
            wrapped.original = args[0]
            pass
        return wrapped
    return wrapper

@original
def datetime(func):
    def wrapper(*args, **kwargs):
        kwargs.update(datetime=dt.utcnow())
        return func(*args, **kwargs)
    return wrapper

@original
def timestamp(func):
    def wrapper(*args, **kwargs):
        utc_timestamp = int(time.mktime(dt.utcnow().timetuple()))
        kwargs.update(timestamp=utc_timestamp)
        return func(*args, **kwargs)
    return wrapper

@original
def validator(func):
    def wrapper(*args, **kwargs):
        controller = args[0]
        try:
            return func(*args, **kwargs)
        except ValidationError as e:
            controller.context.setdefault('validation_errors', {})
            controller.context['validation_errors'].update(e.errors)
            return controller.render()
        pass
    return wrapper

@original
def csrf(func):
    def wrapper(*args, **kwargs):
        controller = args[0]
        utc_timestamp = kwargs.get('timestamp')
        if not utc_timestamp:
            utc_timestamp = int(time.mktime(dt.utcnow().timetuple()))
            pass

        csrf_token = controller.session.get('csrf_token')
        csrf_expires = int(controller.session.get('csrf_expires', 0))
        csrf_token_expired = False
        if not (csrf_token and csrf_expires) or csrf_expires < timestamp:
            if csrf_token:
                csrf_token_expired = True
                pass
            controller.session.update(csrf_token=unique_id(24),
                                      csrf_expires=timestamp + controller.app_csrf_token_lifespan)
            pass
        if controller.post:
            if controller.form.get('csrf_token') != csrf_token:
                if csrf_token_expired:
                    string = 'The form timed out before submission as a security measure. Please try again.'
                    pass
                else:
                    string = 'The form submission token did not match the session. Submision was stopped as a security measure.'
                    pass
                raise ValidationError({controller.model.canonical_name: string})
                pass
            pass

        controller.context.update(csrf_token=csrf_token)
        return func(*args, **kwargs)
    return wrapper



from flask import render_template, jsonify, redirect
from flask import url_for, send_from_directory
from flask import request, session

from highfield import naming

class MetaController(type):
    def __init__(cls, name, bases, clsdict):
        cls.canonical_name = naming.class_to_canonical(name)
        super(MetaController, cls).__init__(name, bases, clsdict)
        pass
    pass

class BaseController(object):
    __metaclass__ = MetaController

    def __init__(self, app, action):
        self.app = app
        self.action = action

        self.context = {}
        self.session = session
        self.cookies = request.cookies
        self.args = request.args
        self.form = request.form

        self.controllers = app.controllers
        self.models = app.models

        self.model = getattr(self.models, self.__class__.canonical_name, None)

        self.head = False
        self.get = False
        self.post = False
        self.put = False
        self.delete = False
        setattr(self, request.method.lower(), True)
        pass

    def uses(self, classname, action=None):
        cls = getattr(self.app.controllers, classname)
        return cls(self.app, action or self.action)

    def render(self, context=None, named_view=None, named_extension=None):
        if not named_view:
            folder_name = self.__class__.canonical_name
            file_name = self.action
            if not named_extension:
                named_extension = 'html'
                pass
            named_view = '%s/%s.%s' % (folder_name, file_name, named_extension)
            pass
        if context is not None:
            self.context.update(context)
        context = self.context
        return render_template(named_view, **context)

    def json(self, dictionary):
        return jsonify(**dictionary)

    def redirect(self, redirect_to, status_code=302):
        return redirect(redirect_to, status_code)

    def static_file(self, path):
        return send_from_directory(self.app.static_folder)

    def route_url(self, route):
        return url_for(route)
    pass
